/**************************** CPPHeaderFile ***************************

* FileName [Learner.h]

* PackageName [lc]

* Synopsis [Header file for Learner class.]

* Description [This class encapsulates a generalized MN learner]

* SeeAlso [learner.cpp]

* Author [Nishant Sinha]

* Copyright [ Copyright (c) 2002 by Carnegie Mellon University.  All
* Rights Reserved. This software is for educational purposes only.
* Permission is given to academic institutions to use, copy, and
* modify this software and its documentation provided that this
* introductory message is not removed, that this software and its
* documentation is used for the institutions' internal research and
* educational purposes, and that no monies are exchanged. No guarantee
* is expressed or implied by the distribution of this code. Send
* bug-reports and/or questions to: nishants+@cs.cmu.edu. ]

**********************************************************************/

#ifndef __LEARNER_H__
#define __LEARNER_H__

#include"teacher.h"
#include"fa.h"

//template <typename Teacher, typename Row_el, typename Col_el, typename Trace, typename DFA>
template <typename Teacher>
class Learner
{
 public:
 
	typedef typename Teacher::_DFA DFA;
	typedef typename DFA::St St;
	typedef typename DFA::Sig Sig;
	typedef typename DFA::Ev Ev;

	typedef typename Teacher::Sigma Sigma;

	typedef typename Teacher::Trace Trace;
	typedef typename Trace::Prefix Row_el;
	typedef typename Row_el::Ext Row_el_ext;
	typedef typename Trace::Suffix Col_el;

	typedef string Bv;
	typedef std::map<Row_el,Bv> row_type;
	typedef typename row_type::iterator row_it;
	typedef typename vector<Col_el>::iterator col_it;

/*  Learner() ;
  Learner(Teacher *t); 

  void display_table();
  void learn_from_ce(const Row_el &ce);
  void compute_closure();
  void add_row_by_mem_q(row_type &RowSorSA,const Row_el &access);
  bool check_closed_change();
  void add_col_by_mem_q(const Col_el &distinguish);
  void compute_candidate();
  DFA *get_new_candidate();
  DFA *get_prev_candidate() const;

  void reset_mat(Teacher *t);
  */
  protected:
  Teacher *teacher;
  DFA candidate;
  row_type rowS;
  row_type rowSA;
  vector<Col_el> cols;

  //map from candidate state to Row_el - needed in learn_from_ce
  map<St,Row_el> state2rowel;
  

  void add_extensions_to_rowsa(const Row_el& S) {
	  const Sigma& sigma = teacher->get_sigma();
	  typename_const_foreach(it, sigma) {
		  Row_el SA(S); SA += Row_el_ext(*it);
		  add_row_by_mem_q(rowSA,SA);
	  }                
  }

  public:

/*********************************************************************/
//constructors and destructors
/*********************************************************************/
Learner() : teacher(NULL), candidate(false) {}

Learner(Teacher *t) : teacher(t), candidate(false)
{
  cols.push_back(Col_el());
  add_row_by_mem_q(rowS, Row_el());
  add_extensions_to_rowsa(Row_el());
  compute_closure();
}


/*********************************************************************/
//given a path ce, and index, simulate the first index symbols of ce
//on the latest conjecture DFA and the remaining symbols on the
//unknown DFA. then return if an accepting state is reached.
/*********************************************************************/
bool compute_alpha(const Trace &ce,const size_t index) const
{
  CHK (index <= ce.size());
  St st;
  candidate.run(ce,index,st);
  typename map<St,Row_el>::const_iterator j = state2rowel.find(st);
  CHK (j != state2rowel.end());
  Col_el suffix; ce.get_suffix(index,suffix);
  Util::msg(3,"[index] "+int2str(index)+", [prefix] "+j->second.to_str()+ ", [suffix] "+suffix.to_str());
  Trace t(j->second,suffix); 
  return teacher->ask_mem_q(t);
}

/*********************************************************************/
//learn a counterexample
/*********************************************************************/
void learn_ce(const Trace &ce)
{
	Util::message(2,"\n####### begin learnCE ...ce = %s\n",ce.to_str().c_str());
	size_t ceLen = ce.size();
	bool posCE = compute_alpha(ce,0); 
	size_t low = 0,high = ceLen;
	bool highCE = compute_alpha(ce,high);
	CHK (posCE != highCE);
	while(low <= high) {    
		int mid = (low+high) / 2;
		bool acceptVal1 = compute_alpha(ce,mid);
		bool acceptVal2 = compute_alpha(ce,mid + 1);
		if(acceptVal1 != acceptVal2) {
			Col_el distinguish; ce.get_suffix(mid + 1,distinguish);
			add_col_by_mem_q(distinguish);
			compute_closure();
			Util::message(2,"####### end learnCE ...\n");
			return;
		}
		if(posCE == acceptVal1) low = mid + 1;
		else high = mid - 1;
	}
	assert (false);
}

/*********************************************************************/
//add a row to the table
/*********************************************************************/
void add_row_by_mem_q(row_type &RowSorSA,const Row_el& access)
{
	Util::msg(1,"add_row_by_mem_q: "+access.to_str()+"\n"); display_table();
	CHK ( RowSorSA.find(access) == RowSorSA.end() );
	Bv row_bv;
	for(size_t i = 0;i< cols.size();++i) {
		Trace member(access, cols[i]);
		bool flag = teacher->ask_mem_q(member);   
		row_bv.append(flag ? "1" : "0");
		RowSorSA[access] = row_bv;
	}
	display_table();
	Util::msg(1,"end add_row_by_mem_q \n"); 
}

/*********************************************************************/
//add a column to the table
/*********************************************************************/
void add_col_by_mem_q(const Col_el &distinguish)
{
  Util::message(3,"distinguish is : %s\n",distinguish.to_str().c_str());
  for(size_t i = 0;i < cols.size();++i) CHK (!(cols[i] == distinguish));
  cols.push_back(distinguish);  
  //add in rowS
  for(row_it it = rowS.begin();it != rowS.end();++it) {
    Trace t(it->first, distinguish);
    bool flag = teacher->ask_mem_q(t);
    it->second.append(flag? "1" : "0");
  }
  //add in rowSA
  for(row_it it = rowSA.begin();it != rowSA.end();++it) {
    Trace t(it->first,distinguish);
    bool flag = teacher->ask_mem_q(t);        
    it->second.append(flag? "1" : "0");
  }
}


/*********************************************************************/
//check if the table has changed from closed to non-closed state
/*********************************************************************/
bool check_closed_change()
{
  Util::message(3,"begin ClosedChange ...\n");
  for(row_it itR = rowSA.begin();itR != rowSA.end();++itR) {    
    const Row_el &SA = itR->first;
    const Bv &bv = itR->second;
    bool isClosed = false;
    for(row_it it1 = rowS.begin();it1 != rowS.end();++it1) {
      if((it1->second == bv)) {
        isClosed = true;
        break;
      }
    }
    //this row "bv" does not exist in rowS 
    if(!isClosed) { 
      Util::message(3,"checkClosedChange : rowS[%s] = %s does not exist ...\n",SA.to_str().c_str(), bv.c_str());
		add_extensions_to_rowsa(SA);
      rowS[SA] = bv;
      rowSA.erase(SA);
      return true;
    }           
  }             
  Util::message(3,"end ClosedChange ...\n");
  return false;
}


/*********************************************************************/
//make the table closed and consistent
/*********************************************************************/
void compute_closure()
{
  Util::message(3,"start compute_closure ...\n");
  bool closedChanged = check_closed_change();
  while(closedChanged) closedChanged = check_closed_change();
  Util::message(3,"end compute_closure ...\n");
}


/*********************************************************************/
//compute the next candidate
/*********************************************************************/
void compute_candidate()
{
    CHK (!check_closed_change());
    //display_table();
    Util::message(2,"begin compute_candidate ...\n");
    Util::msg(3,"Curr table is ..\n"); display_table();
    const Sigma& sigma = teacher->get_sigma();
    Util::message(2,"teacher is %x\n",teacher);
    //if(candidate != NULL) delete candidate;
    //candidate = new DFA(false); //with no sink state
    candidate = DFA(true); //allow sink state
    candidate.set_sigma(sigma);
    state2rowel.clear();
    map<Bv,St> bv2state;

    //make states in DFA corr to rows
    for(row_it it = rowS.begin();it != rowS.end();++it) {
        Bv currBv(it->second);
        CHK (bv2state.find(currBv) == bv2state.end());
        //create state
        St s = candidate.new_state();
        bv2state[currBv] = s;
        state2rowel[s] = it->first;
        //final states
        if(currBv[0] == '1') candidate.set_final(bv2state[currBv]);
    }

    //start state
    CHK (rowS.find(Row_el()) != rowS.end());
    Bv bv_start(rowS[Row_el()]);
    candidate.set_init(bv2state[bv_start]);

    //transitions
    for(row_it itR = rowS.begin();itR != rowS.end();++itR) {
        Bv currBv(itR->second);
        CHK (bv2state.find(currBv) != bv2state.end());
        const St &state1 = bv2state[currBv];    
        typename_const_foreach(it, sigma) {
            Row_el r(itR->first); r += Row_el_ext(*it);              
            if(rowSA.find(r) != rowSA.end()) {
                Bv nextBv(rowSA[r]);
                CHK (bv2state.find(nextBv) != bv2state.end());
                const St &state2 = bv2state[nextBv];
                candidate.add_trans(state1,*it,state2);
            } else if(rowS.find(r) != rowS.end()) {
                Bv nextBv(rowS[r]);
                CHK (bv2state.find(nextBv) != bv2state.end());
                const St &state2 = bv2state[nextBv];
                candidate.add_trans(state1,*it,state2);
            }
            else Util::error(false,"ERROR : %s not found in rowS or rowSA\n",r.to_str().c_str());
        }
    }
    Util::message(2,"new candidate now has %d states\n",candidate.get_num_states()); 
    Util::msg(3,candidate.to_str());
    Util::message(2,"\nend compute_candidate ...\n");
}

/*********************************************************************/
//return the new candidate DFA
/*********************************************************************/
const DFA& get_new_candidate()
{
  compute_candidate();
  return (candidate);
}

/*********************************************************************/
//return the previous candidate
/*********************************************************************/
const DFA& get_prev_candidate() const { return (candidate); }








/*********************************************************************/
//Remove duplicate rows in rowS, either by moving them to rowSA
//or completely removing them
//Note: The cache manager at teacher's end should still have the removed entries
/*********************************************************************/
void compact()
{
	map< Bv,set<Row_el> > dup_entries;
	for(typename row_type::iterator it = rowS.begin();it != rowS.end();++it) {
		dup_entries[it->second].insert(it->first);
	}		
	for(typename map< Bv,set<Row_el> >::iterator it = dup_entries.begin(); it!=dup_entries.end(); ++it) {
		if( it->second.size() == 1) continue;
		else {
			set<Row_el>& curr_entries = it->second;
			//move all dup entries except smallest from rowS to rowSA
			//find smallest trace entry
			typename set<Row_el>::iterator itE = curr_entries.begin();
			Row_el min_trace = *itE;
			while(itE != curr_entries.end()) {
				if(min_trace.size() > itE->size()) min_trace = *itE;
				++itE;
			}
			curr_entries.erase(min_trace);
			//move rest to rowSA	
			for(itE = curr_entries.begin(); itE!= curr_entries.end(); ++itE) {
				rowSA[*itE]= rowS[*itE];
				rowS.erase(*itE);
			}
		}
	}	 	
	set<Row_el> to_eliminate;
	//mark elements in rowSA which extend those in rowS
	for(row_it it = rowSA.begin();it != rowSA.end();++it) 
		to_eliminate.insert(it->first);

	const set<Sig>& sigma = teacher->get_sigma();
	for(row_it it = rowS.begin();it != rowS.end();++it) 
		for(typename set<Sig>::iterator itA = sigma.begin(); itA!= sigma.end(); ++itA) {
				Row_el SAA(it->first); SAA += Row_el_ext(*itA);
				if(rowSA.find(SAA) != rowSA.end()) to_eliminate.erase(SAA);
		}
	//remove rest from rowSA completely   
	for(typename set<Row_el>::iterator it = to_eliminate.begin(); it!= to_eliminate.end(); ++it) {
		rowSA.erase(*it);
	}
}


/*********************************************************************/
//resetting to a new teacher - TODO: note that in the general case, both
//language and alphabet may change and learner must take care of it.
/*********************************************************************/
void reset_mat(Teacher *t) 
{ 
	Util::message(2,"Begin reset_mat...");
  teacher = t; 
  //Validate the Table
  for(size_t i = 0;i< cols.size();++i) {
    const Col_el& distinguish = cols[i];
    //add in rowS
    for(row_it it = rowS.begin();it != rowS.end();++it) {
      Trace t(it->first, distinguish);
      bool flag = teacher->ask_mem_q(t);
      it->second[i]= (flag? '1' : '0');
    }
    //add in rowSA
    for(row_it it = rowSA.begin();it != rowSA.end();++it) {
      Trace t(it->first, distinguish);
      bool flag = teacher->ask_mem_q(t);        
      it->second[i]= (flag? '1' : '0');
    }      
  }	
  //compact the Table
  compact();
  Util::message(2,"End reset_mat...\n");
}

/*********************************************************************/
//display the table
/*********************************************************************/
void display_table()
{
  Util::message(2,"*********************************\n");
  for(col_it it = cols.begin(); it!= cols.end();++it) {
    Util::message(2,"   %s",it->to_str().c_str());
  }
  Util::message(2,"\n");
  for(row_it it = rowS.begin(); it!= rowS.end();++it) {
    Util::message(2,"%s    :%s\n",it->first.to_str().c_str(),it->second.c_str());
  }
  Util::message(2,"\n-------------------------\n");
  for(row_it it = rowSA.begin(); it!= rowSA.end();++it) {
    Util::message(2,"%s    :%s\n",it->first.to_str().c_str(),it->second.c_str());
  }
  Util::message(2,"*********************************\n");
}


};

#endif
